/*
   BUP_C :  Backupt die Harddisc QL-HD

   Autor: Dirk Steinkopf
    Mrz 1992
   Dieses Programm ist nach Prinzipien und Ideen des Backup-Programms
   von Wolfgang Holzinger entstanden.

   V1.09  25. Mai 1992
          Offset-Fehler bei langen Dateien behoben
          30. Mai 1992 fertiggetestet
*/

#include <stdio.h>
#include <qdos.h>
#include <wini.h>
#include <string.h>
#include <assert.h>
#include <fcntl.h>
#include <time.h>

#define NAME    "bup"
#define VERSION "Bup V1.09  5/92 by DStk\n"
#define WHOAMI  _prog_name

char _prog_name[] = NAME;

#define ALIGN(x) ((x) + ((x)&1))
#define min(a,b) ((a) < (b) ? (a) : (b))
#ifdef DEBUG
#  define dprintf(s,x) printf(s,x)
#else
#  define dprintf(s,x)  /* Nix */
#endif
#define MIN_BUFFER 512
#define MAX_SHORT 32766
#define MAX_FNLEN 36
#define MAX_FILELEN (32768*512)
#define FNLEN MAXNAMELEN
#define TYP_UPDATEDATE 0
#define TYP_BACKUPDATE 2

extern char *malloc(long);

void std_err(char *);
void std_err3(char *, char *, char *);
void errout(char *, int);
void usage(void);
char *get_bigmem(long *memlen, int *total);
long io_open_askover(char *fn);

void do_backup(void);
int backup (char *dir, int depth);
void save_hdr(struct qdirect *fh);
void save_file(char *fn, struct qdirect *fh);
void save_mem(char *mem, long totlen, long restlen);
void set_bkdt(int fd);
void format_prompt(void);
void format_disc(void);
void newdisc(long offset);

void do_restore(void);
int restore_file(struct qdirect *fh, char *fn);
int rest_newdisc(void);

char savefile[FNLEN+1] = "flp1_BACKUP";
char backname[FNLEN+1] = "";
char mapfile[FNLEN+1];
char topdir[FNLEN+1];
FILE *mapfp;
long savelen, backchid = 0;
int discnr;
short verb_flag = 0, all_flag = 0, format_flag = 0, map_flag = 0, over_flag = 0;


short restore = 0;
char destdev[FNLEN+1];
char wildcard[FNLEN+1];

struct preamble {       /* Header vor jedem Backup-File */
   char string[14];     /* Identifikationsstring */
   char discnam[10];    /* Name der Disc (= Name der Backup) */
   char discnum[4];     /* Fortlaufende Nummer innherhalb dieser Backup */
   char res[4];         /* reserviert (leer) */
   long date;           /* Datum der Backup */
   long offset;         /* Anz. der Bytes bis zum ersten Fileheader */
};
#define ID_STRING "QL-HD BACKUP"


static char *month[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};



void usage(void)
{
   std_err("Usage:\n");
   std_err("backup: bup [options] [dir]\n");
   std_err("        backup files in dir or current data directory\n");
   std_err("resore: bup -r [options] [[device_]wildcard]\n");
   std_err("        restore all files (to which wildcard applies - if given)\n");

   std_err("Options:\n");
   std_err("   -v       verbose mode on\n");
   std_err("   -V       only print version of bup, then exit\n");
   std_err("   -bfile   save to/restore from file instead of flp1_BACKUP\n");
   std_err("   -nname   backup's name is name (used in map and backup file)\n");
   std_err("   -o       overwrite existing files\n");
   std_err("following options are only useful in backup mode:\n");
   std_err("   -f       format discs before writing to them\n");
   std_err("   -a       backup all files (default: backup files that have changed)\n");
   std_err("   -m[map]  write a map to file map. If map is nullstring write no map\n\n");

   exit(1);
}


void main(int argc, char *argv[])
{
   char *a;

   setbuf(stdout, NULL);
   std_err(VERSION);
   for (argc--, argv++; argc > 0; argc--, argv++)  {
      a = *argv;
      if (a[0] != '-' || a[1] == '\0')
         break;
      while (*++a) {
         switch(*a) {
            case 'v': verb_flag++;
                      break;
            case 'a': all_flag++;
                      break;
            case 'r': restore++;
                      break;
            case 'o': over_flag++;
                      break;
            case 'b': strcpy(savefile, a + 1);
                      a += strlen(savefile);
                      break;
            case 'n': strcpy(backname, a + 1);
                      a += strlen(backname);
                      break;
            case 'm': strcpy(mapfile, a + 1);
                      a += strlen(mapfile);
                      map_flag++;
                      break;
            case 'f': format_flag++;
                      break;
            case 'V': /* std_err(VERSION); */
                      exit(1);
            default:  usage();
         }
      }
   }
   if (restore) {
      if (argc) {
         if (isdevice(argv[0])) {
            /* z.B. "hdk1_cdir_*_h" */
            strncpy(destdev, argv[0], FNLEN);
            *(strchr(destdev, '_') + 1) = '\0';
            strncpy(wildcard, argv[0] + strlen(destdev), FNLEN);
         } else {
            /* z.B. "cdir_*_h" */
            getcwd(destdev, FNLEN);
            *(strchr(destdev, '_') + 1) = '\0';
            strncpy(wildcard, argv[0], FNLEN);
         }
         argc--;
      } else {
         getcwd(destdev, FNLEN);
         *(strchr(destdev, '_') + 1) = '\0';
         strcpy(wildcard, "*");
      }
      /* strcat(destdev, "rest_"); */
   } else {
      if (argc) {
         strncpy(topdir, argv[0], FNLEN);
         argc--;
      } else {
         getcwd(topdir, FNLEN);
         *(strchr(topdir, '_') + 1) = '\0';  /* Ganze Harddisc */
      }
      if (format_flag && !isdevice(savefile))
          errout("Backup file must have a device if you want to format.", 1);
   }
   if (argc)
      usage();

   if (restore)
      do_restore();
   else
      do_backup();

   exit(0);
}



void do_backup(void)
{
   if (backname[0] == '\0') {
      if (all_flag)
         strcpy(backname, "all");
      else
         strcpy(backname, "inc");
   }

   newdisc(0);

   if (backup(topdir, 1))
      fprintf (stderr, "%s: Cannot open directory %s\n", WHOAMI, topdir);

   if (verb_flag)
      printf("%ld bytes saved on %d disc%s.\n",
         savelen, discnr, discnr == 1 ? "" : "s");
   io_close(backchid);
   if (mapfp != NULL)
      fclose(mapfp);
}



int backup (char *dir, int depth)
{
   struct DIR_LIST   *dlist, *first;
   char              dirname[FNLEN+5+1], fname[FNLEN+1];
   int               err;

   assert(stackcheck());
   if (depth == 1 && !(isdevice(dir) && *(strchr(dir, '_') + 1) == '\0')) {
      long fid;
      struct qdirect fh;

      if ( (fid = open_qdir(dir)) < 0)
         return 1;
      if (fs_headr (fid, -1L, &fh, DREADSIZE) == 0)  {
         fprintf (stderr, "%s: Cannot read header !\n", WHOAMI);
         exit(1);
      }
      io_close(fid);
      if (all_flag || fh.d_update > fh.d_backup)
         save_file(dir, &fh);
   }
   strcpy(dirname, dir);
   if (dirname[strlen(dir)-1] != '_')
      strcat(dirname, "_");
   strcat(dirname, "*");
   first = dlist = qdir_read(dirname, NULL, QDR_ALL, &err);
   if ((int)dlist < 0)
      return 1;
   for ( ; dlist != NULL; dlist = dlist->dl_next)  {
      if (all_flag || dlist->dl_dir.d_update > dlist->dl_dir.d_backup)
         save_file(dlist->dl_cname, &dlist->dl_dir);
      if (dlist->dl_dir.d_type == QF_DIR_TYPE)  {
         if (backup(dlist->dl_cname, depth+1))
            fprintf (stderr, "%s: Cannot open directory %s\n",
               WHOAMI, dlist->dl_cname);
      }
   }
   qdir_delete(first);
   return 0;
}



void save_file(char *fn, struct qdirect *fh)
{
   int copyfd, total, err, backup_date;
   long copychid, memlen, l, l_sum, date;
   char *mem, abs_fname[FNLEN+1];
   struct tm *tim;

   if (fh->d_type == QF_DIR_TYPE) {
      if (fh->d_backup != 0 && !all_flag)
         return;
      fh->d_length = 0;
      if (verb_flag)
         puts(fn);
   } else {
      fh->d_length -= DREADSIZE;
      if (verb_flag)
         printf("%-40s %l7d\n", fn, fh->d_length);
   }
   if (mapfp != NULL)
      fprintf(mapfp, "%s %02d %-37s",
         backname, discnr, qlstr_to_c(abs_fname, &fh->d_szname));

   copyfd = open(fn, O_RDONLY);
   if (copyfd < 0) {
      std_err3(WHOAMI, "Cannot open ", fn);
      fh->d_length = 0;
      save_hdr(fh);      /* Dummy header */
      if (mapfp != NULL)
          fprintf(mapfp, " ...cannot open.\n");
      return;
   }
   copychid = getchid(copyfd);

   memlen = fh->d_length;
   if (memlen == 0) {
      save_hdr(fh);
   } else {
      mem = get_bigmem(&memlen, &total);
      if (total) {
         if (memlen != fs_load(copychid, mem, memlen)) {
            std_err3(WHOAMI, "Cannot load ", fn);
            if (mapfp != NULL)
               fprintf(mapfp, " ...cannot load.\n");
            fh->d_length = 0;
            save_hdr(fh);      /* Dummy header */
            free(mem);
            close(copyfd);
            return;
         }
         save_hdr(fh);
         save_mem(mem, memlen, 0);
      } else {
         save_hdr(fh);
         l_sum = 0;
         while( (l = io_fstrg(copychid, -1, mem, memlen)) != 0 ) {
            save_mem(mem, l, fh->d_length - l_sum - l);
            l_sum += l;
         }
         if (l_sum != fh->d_length) {
            std_err3(WHOAMI, fn, " was read incompletely.");
            save_mem(mem, fh->d_length - l_sum, 0); /* Dummy, damit Lnge stimmt */
            free(mem);
            close(copyfd);
            if (mapfp != NULL)
                fprintf(mapfp, " ...incomplete.\n");
            return;
         }
      }
      free(mem);
   }
   set_bkdt(copyfd);
   if (mapfp != NULL) {
      date = TIME_QL_UNIX(fh->d_update);
      tim = localtime(&date);
      fprintf(mapfp, "  %7ld   %04d %s %02d\n",
         fh->d_length, 1900 + tim->tm_year, month[tim->tm_mon], tim->tm_mday);
   }
   close(copyfd);
}


void save_mem(char *mem, long totlen, long remain)
{
   long schon, copylen, restlen, l2, l3,
        align = ALIGN(totlen);

   for(schon = 0; schon < align; schon += copylen, mem += copylen) {
      copylen = min(align - schon, MAX_SHORT);
      l2 = io_sstrg(backchid, -1, mem, copylen);
      if (l2 < copylen) {
         restlen = copylen - l2;
         do {
            newdisc(align + ALIGN(remain) - schon - l2);
            l3 = io_sstrg(backchid, -1, mem + l2, restlen);
            if (l3 != restlen) {
#if 0
               fprintf(stderr, "l3=%ld, restlen=%ld, _oserr=%d   ",
                  l3, restlen, _oserr);
#endif
               errout("Disc error - try again.", 0);
               discnr--;
            }
         } while (l3 != restlen);
      }
   }
   savelen += align;
}



void save_hdr(struct qdirect *fh)
{
   char buf[22];
   struct QLSTR *mednam = (struct QLSTR *)buf;
   int unused, good;

   fs_mdinf(backchid, -1, mednam, &unused, &good);
   if (unused <= 0)
      newdisc(0);
   savelen += sizeof(*fh);
   io_sstrg(backchid, -1, fh, sizeof(*fh));
}



void set_bkdt(int fd)
{
   long backup_date = 0;  /* Backupdatum auf aktuelle Zeit setzen */

   io_fdate(getchid(fd), -1, TYP_BACKUPDATE, &backup_date);
}



void format_prompt(void)
{
   fputs("Insert another disc, then press y and ENTER ", stderr);
   while (toupper(getchar()) != 'Y')
      ;
   while (getchar() != '\n')
      ;
}



void format_disc(void)
{
   long t,g;
   char drive[FNLEN+1];

   strcpy(drive, savefile);
   strcpy(strchr(drive, '_') + 1, "BackupDisc");
   do {
      while (io_format(drive, &t, &g) < 0) {
         errout("Format failed.", 0);
         format_prompt();
      }
      if (t != g) {
         fprintf(stderr, "Total: %d Sectors   Good: %d Sectors\n", t, g);
         format_prompt();
      }
   } while (t != g);
}



void newdisc(long offset)
{
   char buf[40];
   struct preamble p;

   if (backchid != 0) {
      io_close(backchid);
      discnr++;
   } else {  /* Neustart - alles initen */
      savelen = 0;
      discnr = 1;
      if (!map_flag) {
         strcpy(mapfile, "backup");
         /* strcat(mapfile, backname); */
         strcat(mapfile, "_map");
      }
      if (*mapfile != '\0') {
         mapfp = fopen(mapfile, all_flag ? "w" : "a");
         if (mapfp == NULL)
            errout("Cannot open mapfile.", _oserr);
      } else
         mapfp == NULL;
   }
   do {
      fprintf(stderr, "Insert %sdisc %s %02d, then press ENTER ",
         format_flag ? "new " : "", backname, discnr);
      while (getchar() != '\n')
         ;
      if (format_flag)
         format_disc();
      backchid = io_open_askover(savefile);
      if (backchid < 0)
         errout("Cannot open backup file.", 0);
   } while (backchid < 0);

   memset(&p, 0, sizeof(p));
   strcpy(p.string, ID_STRING);
   strncpy(p.discnam, backname, 9);
   sprintf(p.discnum, "%02d", discnr);
   p.date = mt_rclck();
   p.offset = offset;
   io_sstrg(backchid, -1, &p, sizeof(p));
}



void do_restore(void)
{
   struct qdirect fh;
   char fname[FNLEN+1];

   while (rest_newdisc()) {
      while (io_fstrg(backchid, -1, &fh, DREADSIZE) == DREADSIZE) {
         if (fh.d_length >= MAX_FILELEN || fh.d_szname > MAX_FNLEN) {
            errout("Corrupted backup file - abort.", 0);
            break;
         }
         qlstr_to_c(fname, &fh.d_szname);
         if (verb_flag)
            printf("%-40s %7ld", fname, fh.d_length);
         if (!fnmatch(fname, wildcard)) {
            fs_pos(backchid, ALIGN(fh.d_length), SEEK_REL);
            if (verb_flag)
               puts("   not read.");
            continue;
         }
         if (verb_flag)
            puts(" ...");
         if (restore_file(&fh, fname) == 0)
            ;
      }
      io_close(backchid);
      backchid = 0;
   }
}



int restore_file(struct qdirect *fh, char *fn)
{
   int total, version_nr, err;
   long memlen;
   char *mem;
   long copylen, l;
   char destname[FNLEN+1];
   long destchid;

   memlen = min(MAX_SHORT, fh->d_length);
   mem = get_bigmem(&memlen, &total);
   strcpy(destname, destdev);
   strcat(destname, fn);
   if (fh->d_type == QF_DIR_TYPE) {
      destchid = io_open(destname, NEW_OVER);
      if (destchid == ERR_IU)
         return 0;
   } else
      destchid = io_open_askover(destname);
   if (destchid < 0) {
      std_err3(WHOAMI, "Cannot create ", destname);
      fs_pos(backchid, ALIGN(fh->d_length), SEEK_REL);
      if (memlen)
         free(mem);
      return 1;
   }
   copylen = fh->d_length;
   while(copylen > 0) {
      do {
         l = io_fstrg(backchid, -1, mem, min(memlen, copylen));
         err = _oserr;
         if (l == 0 && err == ERR_EF) {
            std_err3(WHOAMI, fn, " is continued on next disc:");
            rest_newdisc();
         }
      } while (l == 0 && err == ERR_EF);
      if (l == 0) {
         std_err3(WHOAMI, "Read error during ", fn);
         break;
      }
      copylen -=l;
      if (l != io_sstrg(destchid, -1, mem, l)) {
         std_err3(WHOAMI, "Write error: ", destname);
         break;
      }
   }
   l = copylen + (ALIGN(fh->d_length) - fh->d_length);
   if (l != 0)
      fs_pos(backchid, l, SEEK_REL);
   version_nr = ((unsigned long)fh->d_refdate & 0xFFFF0000) >> 16;
   fs_heads(destchid, -1, fh, DREADSIZE);
   io_fdate(destchid, -1, TYP_UPDATEDATE, &fh->d_update);
   io_fvers(destchid, -1, &version_nr);
   io_close(destchid);
   if (memlen)
      free(mem);
   return (copylen == 0 ? 0 : 1);
}



int rest_newdisc(void)
{
   int               first, further_disk;
   struct preamble   p;

   if (backchid != 0) {
      io_close(backchid);
      first = 0;
      discnr++;
   } else {
      first = 1;
   }
   for(;;) {
      if (first) {
         if (*backname)
            fprintf(stderr, "%s: Insert a backup disc %s, then press ENTER ",
               WHOAMI, backname);
         else
            fprintf(stderr, "%s: Insert a backup disc, then press ENTER ",
               WHOAMI);
         fprintf(stderr, "(e for end) ");
      } else
         fprintf(stderr, "%s: Insert disc %s %02d, then press ENTER ",
            WHOAMI, backname, discnr);
      further_disk = 1;
      while (getchar() != '\n')
         further_disk = 0;
      if (!further_disk)
         return 0;
      backchid = io_open(savefile, OLD_SHARE);
      if (backchid < 0) {
         errout("Cannot open backup file.", 0);
         continue;
      }
      io_fstrg(backchid, -1, &p, sizeof(p));
      if (_oserr || strcmp(p.string, ID_STRING) != 0) {
         std_err3(WHOAMI, savefile, " is not a backup file.");
         io_close(backchid);
         continue;
      }
      if ( (!first &&
            (strcmp(p.discnam, backname) != 0 || atoi(p.discnum) != discnr))
         || (first && *backname && strcmp(p.discnam, backname) != 0) )  {
         errout("Wrong backup disc - try another one", 0);
         io_close(backchid);
         continue;
      }
      if (first) {
         discnr = atoi(p.discnum);
         strcpy(backname, p.discnam);
         if (p.offset) {
            fs_pos(backchid, p.offset, SEEK_REL);
            if (_oserr && _oserr != ERR_EF) {
               errout("Cannot position backup file.", 0);
               io_close(backchid);
               continue;
            }
         }
      }
      break;
   }
   if (verb_flag) {
      long t = TIME_QL_UNIX(p.date);

      printf("This is backup disc %s %02d. It was created %s",
         backname, discnr, ctime(&t));
   }
   return 1;
}



char *get_bigmem(long *memlen, int *total)
{
   char *mem;

   *total = 1;
   if (*memlen == 0)
      return (char *)total;  /* Dummy */
   for(;;) {
      mem = malloc(*memlen);
      if (mem != NULL)
         break;
      *memlen = min(MAX_SHORT, ALIGN(*memlen/2 + 1));
      *total = 0;
      if (*memlen < MIN_BUFFER)
         errout("Not enough memory.", ERR_OM);
   }
   return mem;
}



long io_open_askover(char *fn)
{
   short overw;
   long  chid;

   if (over_flag)
      return io_open(fn, NEW_OVER);
   chid = io_open(fn, NEW_EXCL);
   if (chid != ERR_EX)
      return chid;
   fprintf(stderr, "%s: %s exists, press only ENTER to overwrite ", WHOAMI, fn);
   overw = 1;
   while (getchar() != '\n')
      overw = 0;
   if (overw)
      chid = io_open(fn, NEW_OVER);
   return chid;
}



void std_err(char *s)
{
   write(2, s, strlen(s));
}



void std_err3(char *s1, char *s2, char *s3)
{
   std_err(s1);
   if (s2 != NULL) {
      std_err(": ");
      std_err(s2);
   }
   if (s3 != NULL)
      std_err(s3);
   write(2, "\n", 1);
}



void errout(char *s, int err)
{
   std_err3(WHOAMI, s, NULL);
   if (err)
      exit(err);
}



int atoi(char *s)
{
   /* Das atoi aus der libc_a verwendet nicht die feste Basis 10, */
   /* sondern interpretiert z.B. 010 als Oktalzahl, also Dez. 8   */
   return (int)strtol(s, NULL, 10);
}
