/*
   CHKDSK_C :  Prft die FAT der Harddisc QL-HD

   Autor: Dirk Steinkopf
    Februar 1990
   September 1991, jetzt verwendbar ab QLHD V2.22
*/

#include <stdio.h>
#include <qdos.h>
#include <wini.h>
#include <qlhd.h>


#define WHOAMI  (_prog_name)

char _prog_name[] = "chkdsk";
char _copyright[] = " 8/92 DStk";
char _version[]   = " V3.06 ";

void consetup_title();
void (*_consetup)() = consetup_title;


short *fat, first_use, err = 0, warn = 0;
int   fat_siz, dflag = 0;
char  *hmap, dev_hd[10];
struct ext_mdinf hdinfo;

int  cl_map      _PROTOTYPE((short, char *));
void check       _PROTOTYPE((char *));
void check_file  _PROTOTYPE((fheader *));
void std_err();
void std_err3();
void errout      _PROTOTYPE((char *));
void rest_map();
extern char *calloc _PROTOTYPE((int, unsigned int));



void usage()
{
   std_err("Usage: chkdsk [-dV] [drive]\n");
   exit(1);
}


main(argc, argv)

int   argc;
char  **argv;
{
   char *a;
   int drive;

   for (argc--, argv++; argc > 0; argc--, argv++)  {
      a = *argv;
      if (a[0] != '-' || a[1] == '\0')
         break;
      while (*++a) {
         switch(*a) {
            case 'd': dflag++;
                      break;
            case 'V': std_err3(_prog_name, _version, _copyright);
                      exit(1);
            default:  usage();
         }
      }
   }
   if (argc) {
      drive = atoi(*argv);
      argc--;
   } else
      drive = 1;
   if (argc)
      usage();

   a = find_hd(drive);
   if (a != NULL) {
      std_err3(WHOAMI, a, NULL);
      exit (1);
   }
   strncpy(dev_hd, devnam, len_devnam);
   *(dev_hd + len_devnam) = drive + '0';
   strcpy(dev_hd + len_devnam + 1, "_");
   fat = phys_def->fat;
   fat_siz = FAT_SIZE;
   first_use = FIRSTUSE;

   if ( (hmap = calloc(fat_siz, sizeof(char))) == NULL )
      { perror(WHOAMI);  exit(1); }

   check("");
   rest_map();
   if (err == 0 && warn == 0) {
      printf("Harddisc QL-HD %s is ok.\n", dev_hd);
      exit(0);
   } else {
      printf("Harddisc %s:\n%d error(s) - %d warning(s)\n",
         dev_hd, err, warn);
      exit(2);
   }
}




void check (tdir)

char *tdir;
{
   long  chid;
   char  dirnam[FNLEN+5+2], fn[FNLEN+5+1];
   static fheader dbuf;

   if (stackreport() < 600)
      { std_err3(WHOAMI, "Stack overflow.", NULL); exit(1); }

   strcpy(dirnam, dev_hd);
   strcat(dirnam, tdir);
   if ( (chid = open_qdir(dirnam)) < 0)
      { std_err3(WHOAMI, "Can't open dir ", dirnam);  exit(1); }
   strcat(dirnam, "*");
   if (*tdir == '\0') {   /* root dir */
      if (io_fxinf(chid, -1L, &hdinfo) != 0)
         { std_err3(WHOAMI,"Can't get extended info about ",dev_hd);  exit(1); }
      if (fs_headr(chid, -1L, (char *)&dbuf, (long)HSIZ) == 0)
         { std_err3(WHOAMI, "Can't read header from ", dev_hd);  exit(1); }
       check_file(&dbuf);
   } else
      puts(tdir);
   while (read_qdir(chid, dirnam, fn, (char *)&dbuf, QDR_ALL) == 1)  {
      if (dbuf.length == 0)
         continue;
      check_file(&dbuf);
      if (dbuf.type == (char)QF_DIR_TYPE)  {
         strncpy (fn, dbuf.name, dbuf.name_sz);
         fn[dbuf.name_sz] = '\0';
         check(fn);
      }
   }
   io_close(chid);
}



void check_file(fh)

fheader *fh;
{
   char fn[FNLEN+1];
   int n_cl = (fh->length + hdinfo.xm_alloc - 1)
              / hdinfo.xm_alloc;
   int fat_n_cl;

   strncpy (fn, fh->name, fh->name_sz);
   fn[fh->name_sz] = '\0';
   fat_n_cl = cl_map(fh->first_cluster, fn);
   if (n_cl != fat_n_cl) {
      printf("FAT chain is %d cluster(s) to %s for <%s>\n",
         n_cl < fat_n_cl ? fat_n_cl - n_cl : n_cl - fat_n_cl,
         n_cl < fat_n_cl ? "long" : "short",
         fn);
   }
}



int cl_map(cluster, fn)

short cluster;
char *fn;
{
   short cl = cluster;
   int n = 0;

   while (cl >= first_use && cl < fat_siz)  {
      if (hmap[cl]) {
         printf("%dx  %d in <%s>.\n", hmap[cl]+1, cl, fn);
         err++;
      }
      hmap[cl]++;
      cl = fat[cl];
      n++;
   }
   if (cl != -1) {
      printf("cut %d in <%s>.\n", cl, fn);
      err++;
   }
   return n;
}



void rest_map()

{
   int i;

   for (i = first_use; i < fat_siz; i++)
      if (hmap[i] == 0 && fat[i] != 0 && fat[i] != -3)  {
         if (dflag) {
            printf("mis %d - deleted\n", i);
            fat[i] = 0;
         } else {
            printf("mis %d - not deleted\n", i);
         }
         warn++;
      }
}



void std_err(s)

char *s;
{
   write(2, s, strlen(s));
}



void std_err3(s1, s2, s3)

char *s1, *s2, *s3;
{
   std_err(s1);
   if (s2 != NULL) {
      std_err(": ");
      std_err(s2);
   }
   if (s3 != NULL)
      std_err(s3);
   write(2, "\n", 1);
}



void errout(s)

char *s;
{
   std_err3(WHOAMI, s, NULL);
   exit(1);
}

