/* SPALT_C :
         Druckt ein Textfile zwei- oder mehrspaltig aus.

   Aufruf:  spalt <FILE [optionen]
   Alle Schalter knnen (fr Defaults) weggelassen werden:
   -nAnzahl der Spalten (2)
   -rRechter Rand in Zeichen der Gre 80 CPL (0)
   -lLinker Rand in Zeichen (0)
   -mZeichenabstand zwischen den Spalten (2)
   -wAbstand zu Kopf-/Fuzeilen (1)
   -zZeilen pro Seite (60)
   -e => elite             -c => condensed         -p => 1/8" Zeilenabst.
   -q => NLQ               -i => italic            -d => double strike
   -s => subscript mode (Zeilenzahl wird angeglichen)
   -fText unten (">fff<")
   -hText oben ("Seite%ppp%%ddd")
   -u => Texte oben/unten je einmal pro Spalte
   -b =>   "       "      nicht Bold -> normal
*/


#include "stdio.h"
#include "getargs.h"
#include "qdos.h"
#include "string.h"
#include "time.h"

#ifdef C68
char _prog_name[] = "spalt V2.02";
#endif


static char *month[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun",
                        "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

#if 0
#include "hdk1_!c_inc_stdio_h"
#include "hdk1_!c_inc_getargs_h"
#include "hdk1_!c_inc_qdos_h"
#include "hdk1_!c_inc_string_h"
#endif

#define TAB_DIST 8

#define lin_addr(l,c) (page_mem + \
                      (l)*(ch_per_col+1) + \
                      (c)*(ch_per_col+1)*lin_text_per_col)
#define FF '\f'
#define MAXLLEN 166
#define PICA_COND_LEN 137
#define ELITE_COND_LEN 160


int   spalt_anz = 1;
int   r_margin = 0;
int   l_margin = 8;
int   space_betw_col = 2;
int   foot_dist = 1, head_dist = 1;
int   lin_per_page = 66;
int   mod_elite = 1,    mod_cond = 0,     mod_nlq = 0;
int   mod_ital = 0,     mod_ds = 0,       mod_sub = 0;
int   mod_8 = 0;
char  *txt_oben = "File: fff  -  Seite ppp";
char  *txt_unten = "ddd";
int   headl_once = 1;
int   headl_bold = 1;

int   tot_line_width = 80;
int   lin_text_per_col, ch_per_col;
char  cstr[2][MAXLLEN+1];  /* oben/unten */

int   ende, page_nr = 0;
char  *page_mem, *infile = "stdin";

ARG      Argtab [] =
{
   { 'n', INTEGER,   &spalt_anz,       "number of columns"     },
   { 'r', INTEGER,   &r_margin,        "right margin"          },
   { 'l', INTEGER,   &l_margin,        "left margin"           },
   { 'm', INTEGER,   &space_betw_col,           "space between columns" },
   { 'w', INTEGER,   &head_dist,         "space after headline " },
   { 'v', INTEGER,   &foot_dist,         "space before footline" },
   { 'z', INTEGER,   &lin_per_page,        "lines per page"        },
   { 'e', BOOLEAN,   &mod_elite,       "elite char mode"       },
   { 'c', BOOLEAN,   &mod_cond,        "condensed char mode"   },
   { 'i', BOOLEAN,   &mod_ital,        "italic char mode"      },
   { 'd', BOOLEAN,   &mod_ds,          "double strike char mode" },
   { 'p', BOOLEAN,   &mod_8,           "1/8\" line feed pitch" },
   { 's', BOOLEAN,   &mod_sub,         "subscript char mode"   },
   { 'h', STRING,    (int *)&txt_oben, "headline"              },
   { 'f', STRING,    (int *)&txt_unten,"footline"              },
   { 'b', BOOLEAN,   &headl_bold,      "head/footline bold"    },
   { 'u', BOOLEAN,   &headl_once,      "head/footline only once"},
};

#define TABSIZE ( sizeof (Argtab) / sizeof (ARG) )



main (argc, argv)

int   argc;
char  **argv;

{
   float l_pp_tmp;
   char  *malloc(), *calloc();
   int   sp, i, col, lin;
   FILE  *my_stdin;

   /* Kommandozeile analysieren --------------------------------- */
   argc = getargs (argc, argv, Argtab, TABSIZE);
   if (argc > 1) {
      if ( (argc--, my_stdin = freopen(argv[1], "r", stdin)) == NULL) {
         fprintf (stderr, "Can't open input file !\n");
         exit (ERR_NF);
      }
      infile = argv[1];
      if (argc > 1) {
         fprintf (stderr, "Too many parameters on input line !\n");
         fprintf (stderr, "Usage: spalt [-options] [file]\n");
         exit (ERR_BL);
      }
   }

/*
   if (mod_elite && mod_cond) {
      fprintf (stderr, "Can't combine elite and condensed char mode !\n");
      exit(ERR_BP);
   }
*/
   /* Drucker einstellen ---------------------------------------- */
   if (mod_elite) {
      printf ("%c%c",27,'M');
      tot_line_width = tot_line_width *6/5;
      r_margin = r_margin *6/5;
      l_margin = l_margin *6/5;
      space_betw_col = space_betw_col *6/5;
   }
   if (mod_cond) {
      putchar(15);
      if (!mod_elite) {
         tot_line_width = tot_line_width *PICA_COND_LEN/80;
         r_margin = r_margin *PICA_COND_LEN/80;
         l_margin = l_margin *PICA_COND_LEN/80;
         space_betw_col = space_betw_col *PICA_COND_LEN/80;
      } else {
         tot_line_width = tot_line_width*5/6 *ELITE_COND_LEN/80;
         r_margin = r_margin*5/6 *ELITE_COND_LEN/80;
         l_margin = l_margin*5/6 *ELITE_COND_LEN/80;
         space_betw_col = space_betw_col*5/6 *ELITE_COND_LEN/80;
      }
   }
   if (mod_ital)  printf ("%c%c",27,'4');
   if (mod_ds)    printf ("%c%c",27,'G');
   if (mod_nlq)   printf ("%c%c%c",27,'x',1);
   l_pp_tmp = lin_per_page;
   if (mod_8) {
      printf ("%c%c",27,'0');
      l_pp_tmp = l_pp_tmp *8.0/6.0;
   }
   if (mod_sub) {
      printf ("%c%c%c",27,'S',1);
      printf ("%c%c%c",27,'3',mod_8 ? 14 : 18);
      l_pp_tmp *= 10.0/6.0;  /* 9-Nadler: 2.0 */
   }
   lin_per_page = l_pp_tmp;
   /* Parameterberechnung --------------------------------------- */
   lin_text_per_col = lin_per_page - (*txt_oben == '\0' ? 0 : 1+head_dist)
                                   - (*txt_unten== '\0' ? 0 : 1+foot_dist);
   if (lin_text_per_col < 1) {
      fprintf(stderr,"Need at least one line to print !\n");
      exit(ERR_BP);
   }
   ch_per_col = (tot_line_width
                 - r_margin - l_margin
                 - space_betw_col*(spalt_anz-1)) / spalt_anz;
   if (ch_per_col < 1) {
      fprintf(stderr,"Need at least one char to print !\n");
      exit(ERR_BP);
   }
   if ( (page_mem = malloc((ch_per_col+1)*lin_text_per_col*spalt_anz+10)) == NULL) {
      fprintf(stderr,"Not enough memory to store one Page !\n");
      exit(ERR_OM);
   }
   printf ("%c%c",27,'D');
   sp = l_margin ? l_margin : l_margin + space_betw_col + ch_per_col;
   for (i = l_margin ? 1:2; i <= spalt_anz; i++) {
      putchar(sp);
      sp += space_betw_col + ch_per_col;
   }
   putchar(0);
   /* Ausdruck -------------------------------------------------- */
   while(!ende && rd_page()) {
      if (*txt_oben != '\0') {
         if (headl_once) {
            if (l_margin) putchar ('\t');
            wr_hdftl(0, 0);
         }
         else for (i = 1; i <= spalt_anz; i++) {
            if (l_margin || i > 1) putchar('\t');
            wr_hdftl(i, 0);
         }
         for (i = 0; i <= head_dist; i++)
            putchar('\n');
      }
      for (lin = 0; lin < lin_text_per_col; lin++) {
         for (col = 0; col < spalt_anz; col++) {
            if (l_margin || col > 0)
               putchar('\t');
            printf("%s",lin_addr(lin,col));
         }
         putchar('\n');
      }
      if (*txt_unten != '\0') {
         for (i = 0; i < foot_dist; i++)
            putchar('\n');
         if (headl_once) {
            if (l_margin) putchar('\t');
            wr_hdftl(0, 1);
         }
         else for (i = 1; i <= spalt_anz; i++) {
            if (l_margin || i > 1) putchar('\t');
            wr_hdftl(i, 1);
         }
      }
      putchar('\f');
   }
   /* Drucker rckstellen ---------------------------------------- */
   if (mod_elite) printf ("%c%c",27,'P');
   if (mod_cond)  printf ("%c",18);
   if (mod_ital)  printf ("%c%c",27,'5');
   if (mod_ds)    printf ("%c%c",27,'H');
   if (mod_nlq)   printf ("%c%c%c",27,'x',0);
   if (mod_8 || mod_sub) printf ("%c%c",27,'2');
   if (mod_sub)   printf ("%c%c",27,'T');
   printf ("%c%c",27,'D');
   for (i = 8; i <= 80; i += 8)  putchar(i);
   putchar(0);
   fprintf(stderr, "%d pages.\n", page_nr);
   exit(0);
}



rd_page()

{
   int   lin = 0, col = 0, cur_pos = 0, ch, l, c;
   char  *c_addr;

   c_addr = lin_addr(lin,col);
   while (1) {
      *c_addr = (ch = my_getchar());
      if (ch == EOF) {
         *c_addr = '\0';
         for (c = col; c < spalt_anz; c++)
            for (l = 0; l < lin_text_per_col; l++)
               if (c > col || l > lin)  *lin_addr(l,c) = '\0';
         break;
      }
      if (ch == FF) {
         *c_addr = '\0';
         for (l = lin+1; l < lin_text_per_col; l++)
            *lin_addr(l,col) = '\0';
         if (++col == spalt_anz)
            break;
         lin = 0;
         cur_pos = 0;
         c_addr = lin_addr(lin,col);
         continue;
      }
      if (ch == '\n')
         cur_pos = ch_per_col - 1;
      else
         c_addr++;
      if (++cur_pos == ch_per_col) {
         *c_addr = '\0';
         cur_pos = 0;
         if (++lin == lin_text_per_col) {
            lin = 0;
            if (++col == spalt_anz)
               break;
         }
         c_addr = lin_addr(lin,col);
      }
   }
   if (ch == EOF)
      ende = 1;
   if (lin == 0 && col == 0)
      return 0;
   else {
      page_nr++;
      return 1;
   }
}



wr_hdftl(col, ou)

int col, ou;

{
   char  *p;

   if (headl_bold)  printf("%c%c",27,'E');
   for (p = (ou == 0 ? txt_oben : txt_unten); *p != '\0'; p++) {
      if (strncmp(p,"ppp",3) == 0) {
         p += 2;
         if (col)
            printf("%d%c", (page_nr-1)*spalt_anz+col,
               ende && col == spalt_anz ? '-' : '+');
         else
            printf("%d%c", page_nr, ende ? '-' : '+');
         continue;
      }
      if (strncmp(p,"ddd",3) == 0) {
         time_t lt = time(NULL);
         struct tm *t = localtime(&lt);

         p += 2;
         printf("%04d %.3s%3d",
            1900+t->tm_year,
            month[t->tm_mon],
            t->tm_mday);
         continue;
      }
      if (strncmp(p,"fff",3) == 0) {
         p += 2;
         printf("%s", infile);
         continue;
      }
      if (*p == '%') {
         putchar(' ');
         continue;
      }
      putchar (*p);
   }
   if (headl_bold)  printf("%c%c",27,'F');
}



my_getchar()
{
   int   c;
   static short tab = 0, pos = 0;

   if (tab == 0) {
      c = getchar();
      if (c == '\t') {
         tab = TAB_DIST-(pos % TAB_DIST);
      } else if (c == '\n')
         pos = -1;
   }
   if (tab) {
      c = ' ';
      tab--;
   }
   pos++;

   return (c);
}
