/*
   TREE_C :  Gibt den Directorybaum aus

   Autor: Dirk Steinkopf
   (C) November 1989, HASPOSE

   Flags:
   -l    length of each file is printed
    d    date   of each file is printed
    s    computes total size of all files
    n    file names are not printed
    o    only directorynames are printed
    p    full pathname of all files is printed
    z    output contains only pathnames (suitable fo zoo)
    f    list no subdirs
    V    print out version number
*/

#include <stdio.h>
#include <qdos.h>
#include <wini.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#define WHOAMI

char _prog_name[] = "tree";
char _copyright[] = " 9/92 DStk";
char _version[]   = " V2.04 ";

void consetup_title();
void (*_consetup)() = consetup_title;

#define CLUST_BYTES     (hdinfo.xm_alloc)
#define TOTAL_CLUSTERS(l)  (((l) + CLUST_BYTES - 1) / CLUST_BYTES)

#define FNLEN 36

short lflag = 0, dflag = 0, sflag = 0, nflag = 1, oflag = 0, pflag = 0,
      zflag = 0, fflag = 0;
short a_tty;
long size;
struct ext_mdinf hdinfo;

void spaces();
void strwrite();



main(argc, argv)

int argc;
char **argv;
{
   short i;
   char  startdir[FNLEN+1];
   char  *s;

   if (argc >= 2 && *argv[1] == '-')  {
      i = 1;
      while (i)  {
         switch (argv[1][i++]) {
         case 'l':   lflag = 1; break;
         case 'd':   dflag = 1; break;
         case 'n':   nflag = 0; break;
         case 'o':   oflag = 1; break;
         case 's':   sflag = 1; size = 0; break;
         case 'p':   pflag = 1; break;
         case 'z':   zflag = 1; break;
         case 'f':   fflag = 1; break;
         case 'V':   version();
         case '\0':  if (i == 1)
                        usage();
                     else
                        i = 0;
                     break;
         default:    usage(); break;
         }
      }
      i = 3;
   } else
      i = 2;
   if (argc > i)
      usage();
   if (argc == i) {
      strcpy(startdir, argv[i-1]);
   } else
      getcwd(startdir, FNLEN);  /* current DATA dir */
   a_tty = (isatty(1) && isatty(0));
   if (a_tty)
      setnbf(stdout);
   if (a_tty && nflag && !zflag && !fflag)
      printf ("%s:\n", startdir);
   if (!nflag)
      pflag = dflag = lflag = 0;
   if (tree(startdir, 1))  {
      fprintf (stderr, WHOAMI ": Cannot open directory\n");
      exit (1);
   }
   if (nflag && !zflag && !fflag)
      printf ("<\n");
   if (sflag && !zflag)
      printf ("Total size = %dKB.\n", size * (CLUST_BYTES / 1024));
   exit(0);
}



tree (dir, depth)

char *dir;
int depth;
{
   struct DIR_LIST *dlist, *first;
   char           dirname[FNLEN+5+1], fname[FNLEN+1];
   short          isadir;
   int            err, fnlen, dlen = strlen(dir);

   assert(stackcheck());
   if (depth == 1) {
      long fid;
      struct qdirect fh;

      if ( (fid = open_qdir(dir)) < 0)
         return 1;
      if (fs_headr (fid, -1L, &fh, DREADSIZE) == 0)  {
         fprintf (stderr, WHOAMI ": Cannot read header !\n");
         exit(1);
      }
      if (io_fxinf(fid, -1L, &hdinfo) < 0) {
         fprintf(stderr, WHOAMI ": Cannot get extended information\n");
         exit(1);
      }
      if (sflag) {
         size += TOTAL_CLUSTERS(fh.d_length);
      }
      io_close(fid);
   }
   strcpy(dirname, dir);
   if (dirname[dlen-1] != '_') {
      strcat(dirname, "_");
      dlen++;
   }
   strcat(dirname, "*");
   first = dlist = qdir_read(dirname, NULL, QDR_ALL, &err);
   if ((int)dlist < 0)
      return 1;
   for ( ; dlist != NULL; dlist = dlist->dl_next)  {
#if 1
      if (a_tty && kbhit(0)) {
         getch();
         printf("-- Aborted --\n");
         exit(1);
      }
#endif
      if (sflag)
         size += TOTAL_CLUSTERS(dlist->dl_dir.d_length);
      isadir = dlist->dl_dir.d_type == QF_DIR_TYPE;
      if ( (!oflag && !zflag) ||
           (!oflag && zflag && !isadir) ||
           (oflag && !zflag && isadir) )  {
         if (nflag) {
            if (pflag || zflag) {
               printf(dlist->dl_cname);
               fnlen = strlen(dlist->dl_cname);
            } else {
               spaces(depth*3);
               printf(dlist->dl_cname + dlen);
               fnlen = depth*3 + strlen(dlist->dl_cname) - dlen;
            }
         }
         if (lflag || dflag)
            spaces(fnlen >= 36 ? 1 : 37 - fnlen);
         if (lflag)
            printf ("%6d", dlist->dl_dir.d_length - DREADSIZE);
         if (dflag)  {
            long t = TIME_QL_UNIX(dlist->dl_dir.d_update);
            char *s = ctime(&t);
            spaces(2);
            strwrite(s, strlen(s)-1);
         }
         if (nflag)
           putchar('\n');
      }
      if (isadir && !fflag) {
         if (tree(dlist->dl_cname, depth+1))
            printf (WHOAMI ": Cannot open directory %s\n", dlist->dl_cname);
      }
   }
   qdir_delete(first);
   return 0;
}


void spaces(n)

int n;
{
   fwrite("                                                  ",
     sizeof(char), n, stdout);
}


void strwrite(s, n)

int n;
char *s;
{
   fwrite(s, sizeof(char), n, stdout);
}



usage()
{
   fprintf(stderr, "Usage: tree [options] [dir]\n");
   fprintf(stderr, "    Options:\n");
   fprintf(stderr, "    -l    length of each file is printed\n");
   fprintf(stderr, "     d    date   of each file is printed\n");
   fprintf(stderr, "     s    computes total size of all files\n");
   fprintf(stderr, "     n    file names are not printed\n");
   fprintf(stderr, "     o    only directorynames are printed\n");
   fprintf(stderr, "     p    full pathname of all files is printed\n");
   fprintf(stderr, "     z    output contains only pathnames (suitable for zoo)\n");
   fprintf(stderr, "     f    fast listing (no subdirectories)\n");
   fprintf(stderr, "     V    print out version number\n");
   exit(1);
}


version()
{
   fprintf (stderr, "%s%s%s\n", _prog_name, _version, _copyright);
   exit(1);
}


